# -*- coding: utf-8 -*-
##############################################################################
#
#    OERPLib
#    Copyright (C) 2013 Sébastien Alix.
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################

import re

from oerplib import error

TPL_MODEL = """<
<table cellborder="0" cellpadding="0" cellspacing="0"
       border="1" bgcolor="{model_bgcolor}" height="100%%">
    <tr>
        <td border="0" bgcolor="{model_bgcolor_title}" align="center" colspan="3">
            <font color="{model_color_title}">{name}</font>
        </td>
    </tr>
    {attrs}
    {relations_r}
</table>>"""

TPL_MODEL_SUBTITLE = """
<tr><td> </td></tr>
<tr>
    <td align="center"
        border="0"
        colspan="3"><font color="{color}">[{title}]</font></td>
</tr>
"""

TPL_MODEL_ATTR = """
<tr>
    <td align="left" border="0">- <font color="{color_name}">{name}</font></td>
    <td align="left" border="0">{flags}</td>
    <td align="left" border="0"> <font color="{color_name}">{type_}</font> </td>
</tr>
"""

TPL_MODEL_REL = """
<tr>
    <td align="left" border="0">- {name}</td>
    <td align="left" border="0">{flags}</td>
    <td align="left" border="0"> <font color="{color_name}">{type_}</font></td>
</tr>
"""


def pattern2regex(pattern):
    """Return a regular expression corresponding to `pattern` (simpler
    representation of the regular expression).
    """
    pattern = "^{0}$".format(pattern.replace('*', '.*'))
    return re.compile(pattern)


def match_in(elt, lst):
    """Return `True` if `elt` is matching one of a pattern in `lst`."""
    for regex in lst:
        if regex.match(elt):
            return True
    return False


class Relations(object):
    """Draw relations between models with `Graphviz`."""
    def __init__(self, oerp, models, maxdepth=1, whitelist=None, blacklist=None,
                 attrs_whitelist=None, attrs_blacklist=None, config=None):
        self.oerp = oerp
        self._models = models
        self._maxdepth = maxdepth
        self._whitelist = [pattern2regex(model) for model in (models)]
        self._whitelist.extend(
            [pattern2regex(model) for model in (whitelist or ['*'])])
        self._blacklist = [pattern2regex(model) for model in (blacklist or [])]
        self._attrs_whitelist = [pattern2regex(model)
                                 for model in (attrs_whitelist or [])]
        self._attrs_blacklist = [pattern2regex(model)
                                 for model in (attrs_blacklist or [])]
        # Configuration options
        self._config = {
            'relation_types': ['many2one', 'one2many', 'many2many'],
            'show_many2many_table': False,
            'color_many2one': '#0E2548',
            'color_one2many': '#008200',
            'color_many2many': '#6E0004',
            'model_root_bgcolor_title': '#A50018',
            'model_bgcolor_title': '#64629C',
            'model_color_title': 'white',
            'model_color_subtitle': '#3E3D60',
            'model_bgcolor': 'white',
            'color_normal': 'black',
            'color_required': 'blue',
            'color_function': '#7D7D7D',
            'space_between_models': 0.25,
        }
        self._config.update(config or {})
        # Store relations between data models:
        self._relations = {}
        self._stack = {'o2m': {}}
        # Build and draw relations for each model
        for model in models:
            self._build_relations(self.oerp.get(model), 0)

    def _build_relations(self, obj, depth):
        """Build all relations of `obj` recursively:
            - many2one
            - one2many (will be bound to the related many2one)
            - many2many (will be bound with the eventual many2many from the
              other side)
        """
        # Stop scanning when the maxdepth is reached, or when the data model
        # has already been scanned
        if obj._name in self._models:
            depth = 0
        if depth > self._maxdepth or obj._name in self._relations:
            return
        # Check the whitelist, then the blacklist
        if obj._name not in self._models:
            if self._whitelist:
                if not match_in(obj._name, self._whitelist):
                    return
            if self._blacklist:
                if match_in(obj._name, self._blacklist):
                    return
        # Only increments depth for data models which are not already scanned
        if obj._name not in self._relations:
            depth += 1
        # Scan relational fields of the data model
        fields = obj.fields_get()
        if obj._name not in self._relations:
            try:
                fields_items = fields.iteritems()
            except Exception:
                fields_items = fields.items()
            self._relations[obj._name] = {
                'relations': {},
                'relations_r': {},  # Recursive relations
                'fields': dict((k, v) for k, v in fields_items
                               if not v.get('relation')),
            }
        for name, data in fields.iteritems():
            if 'relation' in data \
                    and data['type'] in self._config['relation_types']:
                rel = data['relation']
                # where to store the relation?
                store_type = obj._name == rel and 'relations_r' or 'relations'
                # flags
                flags = {
                    'required': data.get('required'),
                    'function': data.get('function'),
                    'fnct_inv': data.get('fnct_inv'),
                    'fnct_search': data.get('fnct_search'),
                }
                # many2one
                if data['type'] == 'many2one':
                    # Check if related one2many fields have been registered
                    # for the current many2one relation
                    o2m_fields = obj._name in self._stack['o2m'] \
                        and rel in self._stack['o2m'][obj._name] \
                        and name in self._stack['o2m'][obj._name][rel] \
                        and self._stack['o2m'][obj._name][rel][name] \
                        or []
                    # Add the field
                    self._relations[obj._name][store_type][name] = {
                        'type': 'many2one',
                        'relation': rel,
                        'name': name,
                        'o2m_fields': o2m_fields,
                    }
                    self._relations[obj._name][store_type][name].update(flags)
                # one2many
                elif data['type'] == 'one2many':
                    # 'relation_field' key may be missing for 'one2many'
                    # generated by 'fields.function'
                    rel_f = data.get('relation_field', None)
                    # If it is a normal o2m field (with a relation field), it
                    # will be attached to its corresponding m2o field
                    if rel_f:
                        # Case where the related m2o field has already been
                        # registered
                        if rel in self._relations \
                                and rel_f in self._relations[rel][store_type]:
                            if name not in self._relations[
                                    rel][store_type][rel_f]:
                                self._relations[
                                    rel][store_type][
                                        rel_f]['o2m_fields'].append(name)
                        # Otherwise, we will process the field later (when the
                        # m2o field will be scanned)
                        else:
                            if rel not in self._stack['o2m']:
                                self._stack['o2m'][rel] = {}
                            if obj._name not in self._stack['o2m'][rel]:
                                self._stack['o2m'][rel][obj._name] = {}
                            if rel_f not in self._stack['o2m'][rel][obj._name]:
                                self._stack['o2m'][rel][obj._name][rel_f] = []
                            self._stack[
                                'o2m'][rel][obj._name][rel_f].append(name)
                    # If the o2m field has no relation field available
                    # (calculated by a function, or a related field) the
                    # relation is stored as a standalone one2many
                    else:
                        self._relations[obj._name][store_type][name] = {
                            'type': 'one2many',
                            'relation': rel,
                            'name': name,
                        }
                        self._relations[obj._name][store_type][name].update(
                            flags)
                # many2many
                elif data['type'] == 'many2many':
                    #rel_columns = data.get('related_columns') \
                    #    or data.get('m2m_join_columns')
                    #rel_columns = rel_columns and tuple(rel_columns) or None
                    self._relations[obj._name][store_type][name] = {
                        'type': 'many2many',
                        'relation': rel,
                        'name': name,
                        'third_table':
                        data.get('third_table') or data.get('m2m_join_table'),
                        'related_columns': None,
                    }
                    self._relations[obj._name][store_type][name].update(flags)
                # Scan relations recursively
                rel_obj = self.oerp.get(rel)
                self._build_relations(rel_obj, depth)

    def make_dot(self):
        """Returns a `pydot.Dot` object representing relations between models.

            >>> graph = oerp.inspect.relations(['res.partner'])
            >>> graph.make_dot()
            <pydot.Dot object at 0x2bb0650>

        See the `pydot <http://code.google.com/p/pydot/>`_ documentation
        for details.
        """
        try:
            import pydot
        except ImportError:
            raise error.InternalError("'pydot' module not found")
        output = pydot.Dot(
            graph_type='digraph', overlap='scalexy', splines='true',
            nodesep=str(self._config['space_between_models']))
        try:
            relation_items = self._relations.iteritems()
        except Exception:
            relation_items = self._relations.items()
        for model, data in relation_items:
            # Generate attributes of the model
            attrs_ok = False
            attrs = []
            if self._attrs_whitelist \
                    and match_in(model, self._attrs_whitelist):
                attrs_ok = True
            if self._attrs_blacklist \
                    and match_in(model, self._attrs_blacklist):
                attrs_ok = False
            if attrs_ok:
                subtitle = TPL_MODEL_SUBTITLE.format(
                    color=self._config['model_color_subtitle'],
                    title="Attributes")
                attrs.append(subtitle)
                try:
                    fields_items = sorted(data['fields'].iteritems())
                except Exception:
                    fields_items = sorted(data['fields'].items())
                for k, v in fields_items:
                    color_name = self._config['color_normal']
                    if v.get('function'):
                        color_name = self._config['color_function']
                    if v.get('fnct_inv'):
                        color_name = self._config['color_normal']
                    #if v.get('required'):
                    #    color_name = self._config['color_required']
                    attr = TPL_MODEL_ATTR.format(
                        name=k, type_=v['type'],
                        color_name=color_name,
                        flags=self._generate_flags_label(v))
                    attrs.append(attr)
            # Generate recursive relations of the model
            relations_r = []
            if data['relations_r']:
                subtitle = TPL_MODEL_SUBTITLE.format(
                    color=self._config['model_color_subtitle'],
                    title="Recursive relations")
                relations_r.append(subtitle)
            try:
                itervalues = data['relations_r'].itervalues()
            except Exception:
                itervalues = data['relations_r'].values()
            for data2 in itervalues:
                label = self._generate_relation_label(data2)
                flags = self._generate_flags_label(data2)
                rel_r = TPL_MODEL_REL.format(
                    name=label, flags=flags,
                    color_name=self._config['color_normal'],
                    type_=data2['type'])
                relations_r.append(rel_r)
            # Generate the layout of the model
            model_bgcolor_title = self._config['model_bgcolor_title']
            if model in self._models:
                model_bgcolor_title = self._config['model_root_bgcolor_title']
            tpl = TPL_MODEL.format(
                model_color_title=self._config['model_color_title'],
                model_bgcolor_title=model_bgcolor_title,
                model_bgcolor=self._config['model_bgcolor'],
                name=model,
                attrs=''.join(attrs),
                relations_r=''.join(relations_r))
            # Add the model to the graph
            node = self._create_node(model, 'relation', tpl)
            output.add_node(node)
            # Draw relations of the model
            try:
                itervalues = data['relations'].itervalues()
            except Exception:
                itervalues = data['relations'].values()
            for data2 in itervalues:
                if data2['relation'] in self._relations:
                    edge = self._create_edge(model, data2['relation'], data2)
                    output.add_edge(edge)
        return output

    def _create_node(self, name, type_, tpl=None):
        """Generate a `pydot.Node` object.
        `type_` can take one of these values: ``relation``, ``m2m_table``.
        If a HTML `tpl` is supplied, it will be used as layout for the node.
        """
        import pydot
        types = {
            'relation': {
                'margin': '0',
                'shape': tpl and 'none' or 'record',
                'label': tpl or name,
            },
            'm2m_table': {
                'margin': '0',
                'shape': tpl and 'none' or 'record',
                'color': self._config['color_many2many'],
                'fontcolor': self._config['color_many2many'],
                'label': tpl or name,
            },
        }
        return pydot.Node(name, **types[type_])

    def _create_edge(self, model1, model2, data):
        """Generate a `pydot.Edge` object, representing a relation between
        `model1` and `model2`.
        """
        import pydot
        label = self._generate_relation_label(data, space=6, on_arrow=True)
        return pydot.Edge(
            model1, model2,
            label=label,
            labeldistance='10.0',
            color=self._config['color_{0}'.format(data['type'])],
            fontcolor=self._config['color_{0}'.format(data['type'])])
            #arrowhead=(data['type'] == 'many2many' and 'none' or 'normal'),

    def _generate_flags_label(self, data):
        """Generate a HTML label for status flags of a field
        described by `data`.
        """
        flags = []
        if data.get('required'):
            flags.append("<font color='{color}'>R</font>".format(
                color=self._config['color_required']))
        if data.get('function'):
            name = data.get('fnct_inv') and "Fw" or "F"
            if data.get('fnct_search'):
                name += "s"
            flags.append("<font color='{color}'>{name}</font>".format(
                name=name, color=self._config['color_function']))
        if flags:
            return " &#91;{0}&#93;".format(' '.join(flags))
        return ""

    def _generate_relation_label(self, data, space=0, on_arrow=False):
        """Generate a HTML label based for the relation described by `data`."""
        name_color = self._config['color_{0}'.format(data['type'])]
        label = "{space}<font color='{color}'>{name}</font>".format(
            color=name_color, name=data['name'], space=' ' * space)
        # many2one arrow
        if data['type'] == 'many2one' and data['o2m_fields']:
            label = "{label} <font color='{color}'>← {o2m}</font>".format(
                label=label,
                color=self._config['color_one2many'],
                o2m=', '.join(data['o2m_fields']))
        # one2many "standalone" arrow
        if data['type'] == 'one2many':
            pass
        # many2many arrow
        if data['type'] == 'many2many':
            m2m_table = ''
            if self._config['show_many2many_table']:
                if data.get('third_table'):
                    m2m_table = '({table})'.format(
                        table=data.get('third_table'))
            label = "{space}<font color='{color}'>{name} {m2m_t}</font>".format(
                color=name_color, name=data['name'],
                m2m_t=m2m_table, space=' ' * space)
        # flags
        if on_arrow:
            label += self._generate_flags_label(data)
        # add space on the right
        label = label + "{space}".format(space=' ' * space)
        # closing tag
        if on_arrow:
            label = "<{label}>".format(label=label)
        return label

    def write(self, *args, **kwargs):
        """Write the resulting graph in a file.
        It is just a wrapper around the :func:`pydot.Dot.write` method
        (see the `pydot <http://code.google.com/p/pydot/>`_ documentation for
        details).  Below a common way to use it::

            >>> graph = oerp.inspect.relations(['res.partner'])
            >>> graph.write('relations_res_partner.png', format='png')

        About supported formats, consult the
        `Graphviz documentation <http://www.graphviz.org/doc/info/output.html>`_.
        """
        output = self.make_dot()
        return output.write(*args, **kwargs)

# vim:expandtab:smartindent:tabstop=4:softtabstop=4:shiftwidth=4:



# -*- coding: utf-8 -*-
##############################################################################
#
#    OERPLib
#    Copyright (C) 2013 Sébastien Alix.
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################
"""Implements the :class:`Dependencies` class used to compute dependencies
between server modules.
"""
import copy

from oerplib import error
from oerplib.tools import v

TPL_MODULE = """<
<table cellborder="0" cellpadding="0" cellspacing="0"
       border="1" bgcolor="{module_bgcolor}" height="100%%">
    <tr>
        <td border="0" bgcolor="{module_bgcolor_title}" align="center">
            <font color="{module_color_title}">{name}</font>
        </td>
    </tr>
    {models}
    {comment}
</table>>"""

TPL_MODULE_MODEL = """
<tr>
    <td align="left" border="0">- <font color="{color_model}">{model}</font></td>
</tr>
"""

TPL_MODULE_COMMENT = """
<tr>
    <td align="center" border="0"><font color="{module_color_comment}">{comment}</font></td>
</tr>
"""


def pattern2oerp(pattern):
    """Return a SQL expression corresponding to `pattern` (simpler
    representation of the SQL string matching).
    """
    return pattern.replace('*', '%')


class Dependencies(object):
    """Draw dependencies between modules. Models can be displayed in their
    respecting modules as well.
    """
    def __init__(self, oerp, modules=None, models=None, models_blacklist=None,
                 restrict=False, config=None):
        self.oerp = oerp
        self._restrict = restrict
        self._root_modules = modules or []
        # Configuration options
        self._config = {
            'module_uninst_bgcolor_title': '#DEDFDE',
            'module_uninst_color_title': 'black',
            'module_inst_bgcolor_title': '#64629C',
            'module_inst_color_title': 'white',
            'module_root_bgcolor_title': '#A50018',
            'module_root_color_title': 'white',
            'module_highlight_bgcolor_title': '#1F931F',
            'module_highlight_color_title': 'white',
            'module_bgcolor': 'white',
            'module_color_comment': 'grey',
            'model_color_normal': 'black',
            'model_color_transient': '#7D7D7D',
            'show_module_inst': True,
            'show_module_uninst': False,
            'show_model_normal': True,
            'show_model_transient': False,
        }
        self._config.update(config or {})
        # Check if root modules exist
        self._check_root_modules()
        # List of data models
        self._models = self._get_models_data(
            models or [], models_blacklist or [])
        # List of modules computed according to the `restrict` parameter
        # (display all modules or only modules related to data models)
        self._modules, self._modules_full = self._get_modules(
            self._models, keep=not bool(self._root_modules))
        # Fetch dependencies between modules
        self._scan_module_dependencies()

    #@property
    #def models(self):
    #    """Returns a dictionary of all models used to draw the graph."""
    #    return self._models

    #@property
    #def modules(self):
    #    """Returns a dictionary of all modules used to draw the graph."""
    #    return self._modules

    def _check_root_modules(self):
        """Check if `root` modules exist, raise an error if not."""
        module_obj = self.oerp.get('ir.module.module')
        for module in self._root_modules:
            if not module_obj.search([('name', 'ilike', module)]):
                raise error.InternalError(
                    "'{0}' module does not exist".format(module))

    def _get_models_data(self, models, models_blacklist):
        """Returns a dictionary `{MODEL: DATA, ...}` of models corresponding to
        `models - models_blacklist` patterns (whitelist substracted
        by a blacklist).
        """
        res = {}
        # OpenERP v5 does not have the 'modules' field on 'ir.model' used to
        # bound a data model and its related modules.
        if v(self.oerp.version) <= v('6.0'):
            return res
        models_patterns = \
            [pattern2oerp(model) for model in (models)]
        models_blacklist_patterns = \
            [pattern2oerp(model) for model in (models_blacklist)]
        if models:
            model_obj = self.oerp.get('ir.model')
            args = [('model', '=ilike', model)
                    for model in models_patterns]
            for _ in range(len(args) - 1):
                args.insert(0, '|')
            for model in models_blacklist_patterns:
                args.append('!')
                args.append(('model', '=ilike', model))
            ids = model_obj.search(args)
            for data in model_obj.read(ids, ['model', 'modules', 'osv_memory']):
                if not self._config['show_model_transient'] \
                        and data['osv_memory']:
                    continue
                if not self._config['show_model_normal'] \
                        and not data['osv_memory']:
                    continue
                res[data['model']] = {
                    'model': data['model'],
                    'modules': data['modules']
                    and data['modules'].split(', ') or [],
                    'transient': data['osv_memory'],
                }
        return res

    def _get_modules(self, models=None, keep=False):
        """Returns a dictionary `{MODULE: DATA, ...}` with all modules installed
        (`restrict=False`) or only with modules related to data models
        (`restrict=True`).
        """
        if models is None:
            models = {}
        modules = {}
        modules_full = {}
        # Fetch all the modules installed on the server
        module_obj = self.oerp.get('ir.module.module')
        states_inst = ['installed', 'to upgrade', 'to remove']
        states_uninst = ['uninstalled', 'uninstallable', 'to install']
        states = []
        if self._config['show_module_inst'] \
                and not self._config['show_module_uninst']:
            states = states_inst[:]
        elif not self._config['show_module_inst'] \
                and self._config['show_module_uninst']:
            states = states_uninst[:]
        elif self._config['show_module_inst'] \
                and self._config['show_module_uninst']:
            states = []
        args = states and [('state', 'in', states)] or []
        module_ids = module_obj.search(args)
        for data in module_obj.read(module_ids, ['name', 'state']):
            modules_full[data['name']] = {
                'models': [],
                'depends': [],
                'keep': keep,
                'installed': data['state'] in states_inst
            }
        # Dispatch data models in their related modules
        try:
            model_items = models.iteritems()
        except Exception:
            model_items = models.items()
        for model, data in model_items:
            for module in data['modules']:
                if module in modules_full \
                        and model not in modules_full[module]['models']:
                    modules_full[module]['models'].append(model)
        # Compute the list of modules related to data models
        if self._restrict:
            try:
                model_items = models.iteritems()
            except Exception:
                model_items = models.items()
            for model, data in model_items:
                for module in data['modules']:
                    if module not in modules:
                        modules[module] = {
                            'models': [],
                            'depends': [],
                            'keep': keep,
                            'installed': modules_full[module]['installed'],
                        }
                    if model not in modules[module]['models']:
                        modules[module]['models'].append(model)
            # Root modules are included by default, even if they don't contain
            # any of the matching models
            for module in self._root_modules:
                if module not in modules:
                    modules[module] = {
                        'models': [],
                        'depends': [],
                        'keep': keep,
                        'installed': modules_full[module]['installed'],
                    }
        # Otherwise, just take the full list of modules
        else:
            modules = copy.deepcopy(modules_full)
        return modules, modules_full

    def _scan_module_dependencies(self):
        """Scan dependencies of modules, until reaching each node in
        `root_modules`.  If `root_modules` is empty, dependencies of all
        installed modules will be computed.
        """
        module_obj = self.oerp.get('ir.module.module')
        # Compute dependencies of all installed modules
        for name in self._modules_full:
            module_ids = module_obj.search([('name', '=', name)])
            module = module_obj.browse(module_ids[0])
            for dependency in module.dependencies_id:
                if dependency.name in self._modules_full:
                    self._modules_full[name]['depends'].append(dependency.name)
                if name in self._modules and dependency.name in self._modules:
                    self._modules[name]['depends'].append(dependency.name)
        # In restrict mode, fix modules similar to "root" module (with no direct
        # dependency) while they may have indirect dependencies
        if self._restrict:
            for name, data in list(self._modules.items()):  # Avoid iter modification
                # Detect fake "root" module
                if not data['depends'] and self._modules_full[name]['depends']:
                    self._fix_fake_root_module(name)
        # Mark modules to keep in the graph if they belong to a path
        # leading to one of the root modules
        for module in self._modules:
            queue = []
            queue.append(module)
            # If the module is a root module, we keep it regardless of its
            # dependencies
            if module in self._root_modules:
                self._modules[module]['keep'] = True
            # Recursive function to scan the graph and keep modules
            def process_keep(queue, module):
                for depend in self._modules[module]['depends']:
                    queue.append(depend)
                    # Found? Keep modules concerned by this path
                    if depend in self._root_modules:
                        for mod in queue:
                            self._modules[mod]['keep'] = True
                        break
                    else:
                        process_keep(queue, depend)
                queue.pop()
            process_keep(queue, module)

    def _fix_fake_root_module(self, module):
        """Fix the fake root `module` by finding its indirect dependencies."""
        known_paths = []

        def find_path(path, mod, common_model):
            """Try to found a path from the module `mod` among all installed
            modules to reach any 'restricted' module.
            """
            if set(path) not in known_paths:
                known_paths.append(set(path))
            path.append(mod)
            for depend in self._modules_full[mod]['depends']:
                path.append(depend)
                if set(path) in known_paths:
                    continue
                if depend in self._modules:
                    if common_model:
                        # Has the 'head' module a common data model with
                        # the 'tail' one?
                        mod_tail = self._modules[module]['models']
                        mod_head = self._modules[depend]['models']
                        if list(set(mod_tail) & set(mod_head)):
                            return True
                    else:
                        return True
                path.pop()
                res = find_path(path, depend, common_model)
                if res:
                    return res
            path.pop()
            return False

        path = []
        # Modules in the path should preferably have a common data model
        found_ok = find_path(path, module, common_model=True)
        # If not, we try again without the rule of the common model
        if not found_ok:
            known_paths = []
            path = []
            found_ok = find_path(path, module, common_model=False)
        # Update the graph by adding required modules to satisfy the
        # indirect dependency
        if found_ok:
            for index, mod in enumerate(path):
                if mod not in self._modules:
                    # Add the required module, but without its dependencies
                    self._modules[mod] = copy.deepcopy(self._modules_full[mod])
                    for depend in self._modules[mod]['depends'][:]:
                        if depend not in self._modules:
                            self._modules[mod]['depends'].remove(depend)
                    self._modules[mod]['comment'] = \
                        "Indirect dependency"
                    # Add the current module as a dependency to the previous one
                    previous_mod = path[index - 1]
                    if mod not in self._modules[previous_mod]['depends']:
                        self._modules[previous_mod]['depends'].append(mod)

    @staticmethod
    def _draw_graph_node(module, tpl):
        """Generates a Graphviz node named `module`."""
        import pydot
        return pydot.Node(module, margin='0', shape='none', label=tpl)

    @staticmethod
    def _draw_graph_edge(parent, child):
        """Generates a Graphviz edge between `parent` and `child` modules."""
        import pydot
        return pydot.Edge( parent, child, dir='back')

    def make_dot(self):
        """Returns a `pydot.Dot` object representing dependencies
        between modules.

            >>> graph = oerp.inspect.dependencies(['base'], ['res.partner'])
            >>> graph.make_dot()
            <pydot.Dot object at 0x2f01990>

        See the `pydot <http://code.google.com/p/pydot/>`_ documentation
        for details.
        """
        try:
            import pydot
        except ImportError:
            raise error.InternalError("'pydot' module not found")
        output = pydot.Dot()

        def get_template(module, data):
            """Generate the layout of the module."""
            root = all(not self._modules[depend]['keep']
                       for depend in data['depends'])
            # Model lines
            tpl_models = []
            for model in sorted(data['models']):
                color_model = self._config['model_color_normal']
                if self._models[model]['transient']:
                    color_model = self._config['model_color_transient']
                tpl_models.append(
                    TPL_MODULE_MODEL.format(
                        color_model=color_model, model=model))
            # Module comment
            tpl_comment = None
            if data.get('comment'):
                tpl_comment = "<tr><td> </td></tr>"
                tpl_comment += TPL_MODULE_COMMENT.format(
                    module_color_comment=self._config['module_color_comment'],
                    comment=data['comment'])
            # Module
            module_color_title = self._config['module_inst_color_title']
            module_bgcolor_title = self._config['module_inst_bgcolor_title']
            if root:
                module_color_title = self._config['module_root_color_title']
                module_bgcolor_title = self._config['module_root_bgcolor_title']
            if not root and tpl_models:
                module_color_title = \
                    self._config['module_highlight_color_title']
                module_bgcolor_title = \
                    self._config['module_highlight_bgcolor_title']
            if not data.get('installed'):
                module_color_title = self._config['module_uninst_color_title']
                module_bgcolor_title = \
                    self._config['module_uninst_bgcolor_title']
            tpl = TPL_MODULE.format(
                module_color_title=module_color_title,
                module_bgcolor_title=module_bgcolor_title,
                module_bgcolor=self._config['module_bgcolor'],
                name=module.lower(),
                models=''.join(tpl_models),
                comment=tpl_comment or '')
            return tpl

        try:
            module_items = self._modules.iteritems()
        except Exception:
            module_items = self._modules.items()
        for module, data in module_items:
            if not data['keep']:
                continue
            # Add the module as node
            tpl = get_template(module, data)
            node = self._draw_graph_node(module, tpl)
            output.add_node(node)
            for dependency in data['depends']:
                if not self._modules[dependency]['keep']:
                    continue
                # Add edge between the module and it's dependency
                edge = self._draw_graph_edge(dependency, module)
                output.add_edge(edge)

        return output

    def write(self, *args, **kwargs):
        """Write the resulting graph in a file.
        It is just a wrapper around the :func:`pydot.Dot.write` method
        (see the `pydot <http://code.google.com/p/pydot/>`_ documentation for
        details).  Below a common way to use it::

            >>> graph = oerp.inspect.dependencies(['base'], ['res.partner'])
            >>> graph.write('dependencies_res_partner.png', format='png')
        
        About supported formats, consult the
        `Graphviz documentation <http://www.graphviz.org/doc/info/output.html>`_.
        """
        output = self.make_dot()
        return output.write(*args, **kwargs)

# vim:expandtab:smartindent:tabstop=4:softtabstop=4:shiftwidth=4:



# -*- coding: UTF-8 -*-
##############################################################################
#
#    OERPLib
#    Copyright (C) 2011-2013 Sébastien Alix.
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################
"""Provide the :class:`Model` class which allow to access dynamically to all
methods proposed by a data model."""

import sys  # to check Python version at runtime
import collections

from oerplib.tools import v
from oerplib import error
from oerplib.service.osv import fields, browse

try:
    unicode
except NameError:
    unicode = str

class Model(object):
    """.. versionadded:: 0.5

    Represent a data model.

    .. note::
        This class have to be used through the :func:`oerplib.OERP.get`
        method.

    >>> import oerplib
    >>> oerp = oerplib.OERP('localhost')
    >>> user = oerp.login('admin', 'passwd', 'database')
    >>> user_obj = oerp.get('res.users')
    >>> user_obj
    <oerplib.service.osv.osv.Model object at 0xb75ba4ac>
    >>> user_obj.name_get(user.id) # Use any methods from the model instance
    [[1, 'Administrator']]

    .. warning::

        The only method implemented in this class is ``browse``. Except this
        one, method calls are purely dynamic. As long as you know the signature
        of the model method targeted, you will be able to use it
        (see the :ref:`tutorial <tutorials-execute-queries>`).

    """

    fields_reserved = ['id', '__oerp__', '__osv__', '__data__']

    def __init__(self, oerp, model):
        super(Model, self).__init__()
        self._oerp = oerp
        self._name = model
        self._browse_class = self._generate_browse_class()

    def browse(self, ids, context=None):
        """Browse one or several records (if `ids` is a list of IDs)
        from `model`. The fields and values for such objects are generated
        dynamically.

        >>> oerp.get('res.partner').browse(1)
        browse_record(res.partner, 1)

        >>> [partner.name for partner in oerp.get('res.partner').browse([1, 2])]
        [u'Your Company', u'ASUStek']

        A list of data types used by ``browse_record`` fields are
        available :ref:`here <fields>`.

        :return: a ``browse_record`` instance
        :return: a generator to iterate on ``browse_record`` instances
        :raise: :class:`oerplib.error.RPCError`

        """
        if isinstance(ids, list):
            return browse.BrowseRecordIterator(self, ids, context=context)
            #return browse.BrowseRecordIterator(
            #    model=self,
            #    ids=ids,
            #    context=context)
        else:
            obj = self._browse_class(ids)
            self._refresh(obj, context)
            return obj
            #return self.browse(ids, context)

    def _generate_browse_class(self):
        """Generate a class with all its fields corresponding to
        the model name supplied and return them.

        """
        # Retrieve server fields info and generate corresponding local fields
        fields_get = self._oerp.execute(self._name, 'fields_get')
        cls_name = self._name.replace('.', '_')
        # Encode the class name for the Python2 'type()' function.
        # No need to do this for Python3.
        if type(cls_name) == unicode and sys.version_info < (3,):
            cls_name = cls_name.encode('utf-8')
        cls_fields = {}
        for field_name, field_data in list(fields_get.items()):
            if field_name not in Model.fields_reserved:
                cls_fields[field_name] = fields.generate_field(
                    self, field_name, field_data)
        # Case where no field 'name' exists, we generate one (which will be
        # in readonly mode) in purpose to be filled with the 'name_get' method
        if 'name' not in cls_fields:
            field_data = {'type': 'text', 'string': 'Name', 'readonly': True}
            cls_fields['name'] = fields.generate_field(self, 'name', field_data)

        cls = type(cls_name, (browse.BrowseRecord,), {})
        cls.__oerp__ = self._oerp
        cls.__osv__ = {'name': self._name, 'columns': cls_fields}
        slots = ['__oerp__', '__osv__', '__dict__', '__data__']
        slots.extend(list(cls_fields.keys()))
        cls.__slots__ = slots
        return cls

    def _write_record(self, obj, context=None):
        """Send values of fields updated to the server."""
        context = context or self._oerp.context
        obj_data = obj.__data__
        vals = {}
        for field_name in obj_data['updated_values']:
            if field_name in obj_data['raw_data']:
                field = self._browse_class.__osv__['columns'][field_name]
                field_value = obj.__data__['updated_values'][field_name]
                # Many2One fields
                if isinstance(field, fields.Many2OneField):
                    vals[field_name] = field_value and field_value[0]
                # All other fields
                else:
                    vals[field_name] = field_value
        try:
            if v(self._oerp.version) < v('6.1'):
                res = self.write([obj.id], vals, context)
            else:
                res = self.write([obj.id], vals, context=context)
        except error.Error as exc:
            raise exc
        else:
            # Update raw_data dictionary
            # FIXME: make it optional to avoid a RPC request?
            self._refresh(obj, context)
            return res

    def _refresh(self, obj, context=None):
        """Retrieve field values from the server.
        May be used to restore the original values
        in the purpose to cancel all changes made.

        """
        context = context or self._oerp.context
        obj_data = obj.__data__
        obj_data['context'] = context
        # Get basic fields (no relational ones)
        basic_fields = []
        try:
            column_items = obj.__osv__['columns'].iteritems()
        except Exception:
            column_items = obj.__class__.__osv__['columns'].items()
        for field_name, field in column_items:
            if not getattr(field, 'relation', False):
                basic_fields.append(field_name)
            else:
                obj_data['raw_data'][field_name] = None
        # Fill fields with values of the record
        if obj.id:
            if v(self._oerp.version) < v('6.1'):
                data = self.read([obj.id], basic_fields, context)
                if data:
                    obj_data['raw_data'].update(data[0])
                else:
                    obj_data['raw_data'] = False
            else:
                data = self.read([obj.id], basic_fields, context=context)
                if data:
                    obj_data['raw_data'].update(data[0])
                else:
                    obj_data['raw_data'] = False
            if obj_data['raw_data'] is False:
                raise error.RPCError(
                    "There is no '{model}' record with ID {obj_id}.".format(
                        model=obj.__class__.__osv__['name'], obj_id=obj.id))
        # No ID: fields filled with default values
        else:
            if v(self._oerp.version) < v('6.1'):
                default_get = self.default_get(
                    list(obj.__osv__['columns'].keys(), context))
            else:
                default_get = self.default_get(
                    list(obj.__osv__['columns'].keys(), context=context))
            obj_data['raw_data'] = {}
            for field_name in obj.__osv__['columns']:
                obj_data['raw_data'][field_name] = False
            obj_data['raw_data'].update(default_get)
        self._reset(obj)

    def _reset(self, obj):
        """Cancel all changes by restoring field values with original values
        obtained during the last refresh (object instanciation or
        last call to _refresh() method).

        """
        obj_data = obj.__data__
        obj_data['updated_values'] = {}
        # Load fields and their values
        for field in list(self._browse_class.__osv__['columns'].values()):
            if field.name in obj_data['raw_data']:
                obj_data['values'][field.name] = \
                    obj_data['raw_data'][field.name]
                setattr(obj.__class__, field.name, field)

    def _unlink_record(self, obj, context=None):
        """Delete the object from the server."""
        context = context or self._oerp.context
        if v(self._oerp.version) < v('6.1'):
            return self.unlink([obj.id], context)
        else:
            return self.unlink([obj.id], context=context)

    def __getattr__(self, method):
        """Provide a dynamic access to a RPC method."""
        def rpc_method(*args, **kwargs):
            """Return the result of the RPC request."""
            if v(self._oerp.version) < v('6.1'):
                if kwargs:
                    raise error.RPCError(
                        "Named parameters are not supported by the version "
                        "of this server.")
                result = self._oerp.execute(
                    self._browse_class.__osv__['name'], method, *args)
            else:
                if self._oerp.config['auto_context'] \
                        and 'context' not in kwargs:
                    kwargs['context'] = self._oerp.context
                result = self._oerp.execute_kw(
                    self._browse_class.__osv__['name'], method, args, kwargs)
            return result
        return rpc_method

    def __repr__(self):
        return "Model(%r)" % (self._browse_class.__osv__['name'])

    # ---------------------------- #
    # -- MutableMapping methods -- #
    # ---------------------------- #

    def __getitem__(self, obj_id):
        return self.browse(obj_id)

    def __iter__(self):
        ids = self.search([])
        return browse.BrowseRecordIterator(self, ids)

    def __len__(self):
        return self._oerp.search(self._browse_class.__osv__['name'], count=True)

# vim:expandtab:smartindent:tabstop=4:softtabstop=4:shiftwidth=4:



# -*- coding: UTF-8 -*-
##############################################################################
#
#    OERPLib
#    Copyright (C) 2013 Sébastien Alix.
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published
#    by the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################
"""Provides the :func:`scan_on_change` function."""
import xml.etree.ElementTree
import re

ON_CHANGE_RE = re.compile('^(.*?)\((.*)\)$')


def scan_on_change(oerp, models):
    """Scan all `on_change` methods detected among views of `models`."""
    result = {}
    view_obj = oerp.get('ir.ui.view')
    model_data_obj = oerp.get('ir.model.data')
    for model in models:
        # Get all model views
        view_ids = view_obj.search(
            [('model', '=', model), ('type', 'in', ['form', 'tree'])])
        model_data_ids = model_data_obj.search(
            [('res_id', 'in', view_ids), ('model', '=', 'ir.ui.view')])
        model_data = model_data_obj.read(
            model_data_ids, ['name', 'module', 'res_id'])
        for data in model_data:
            # For each view, find all `on_change` methods
            view_name = "{0}.{1}".format(data['module'], data['name'])
            view_data = oerp.execute(
                model, 'fields_view_get', data['res_id'], 'form')
            _scan_view(model, view_name, view_data, result)
    return result


def _scan_view(model, view_name, view_data, result):
    """Update `result` with all `on_change` methods detected
    on the view described by `view_data`.
    """
    if model not in result:
        result[model] = {}
    # Scan the main view description
    xml_root = xml.etree.ElementTree.fromstring(view_data['arch'])
    # NOTE: Python 2.6 does not support full XPath, it is
    # why the ".//field" pattern is used instead of ".//field[@on_change]"
    for elt in xml_root.findall(".//field"):
        if 'on_change' not in elt.attrib:
            continue
        match = ON_CHANGE_RE.match(elt.attrib['on_change'])
        if match:
            func = match.group(1)
            args = [arg.strip() for arg in match.group(2).split(',')]
            field = elt.attrib['name']
            if func not in result[model]:
                result[model][func] = {}
            if view_name not in result[model][func]:
                result[model][func][view_name] = {}
            if field not in result[model][func][view_name]:
                result[model][func][view_name][field] = []
            if args and args not in result[model][func][view_name][field]:
                args = list(map(_clean_arg, args))
                result[model][func][view_name][field] = args
    # Scan recursively all other sub-descriptions defined in the view
    try:
        fields_items = view_data['fields'].iteritems()
    except Exception:
        fields_items = view_data['fields'].items()
    for field_name, field_data in fields_items:
        if field_data.get('views') and field_data['views'].get('form'):
            model = field_data['relation']
            if field_data['views'].get('form'):
                _scan_view(
                    model, view_name, field_data['views']['form'], result)
            if field_data['views'].get('tree'):
                _scan_view(
                    model, view_name, field_data['views']['tree'], result)
    return result


def _clean_arg(arg):
    return {
        'False': False,
        'True': True,
        'None': None,
    }.get(arg, arg)

# vim:expandtab:smartindent:tabstop=4:softtabstop=4:shiftwidth=4:



# -*- coding: UTF-8 -*-

try:
    import unittest2 as unittest
except:
    import unittest
from datetime import datetime

from args import ARGS

import oerplib


class TestDB(unittest.TestCase):

    def setUp(self):
        self.oerp = oerplib.OERP(
            ARGS.server, protocol=ARGS.protocol, port=ARGS.port,
            version=ARGS.version)
        self.databases = []

    def test_db_list(self):
        res = self.oerp.db.list()
        self.assertIsInstance(res, list)

    def test_db_list_lang(self):
        res = self.oerp.db.list_lang()
        self.assertIsInstance(res, list)

    def test_db_dump(self):
        dump = self.oerp.db.dump(ARGS.super_admin_passwd, ARGS.database)
        self.assertIsNotNone(dump)

    def test_db_restore_new_database(self):
        dump = self.oerp.db.dump(ARGS.super_admin_passwd, ARGS.database)
        date = datetime.strftime(datetime.today(), '%Y-%m-%d_%Hh%Mm%S')
        new_database = "%s_%s" % (ARGS.database, date)
        self.oerp.db.restore(ARGS.super_admin_passwd, new_database, dump)
        self.databases.append(new_database)

    def test_db_restore_existing_database(self):
        dump = self.oerp.db.dump(ARGS.super_admin_passwd, ARGS.database)
        self.assertRaises(
            oerplib.error.RPCError,
            self.oerp.db.restore,
            ARGS.super_admin_passwd, ARGS.database, dump)

    def tearDown(self):
        for db in self.databases:
            try:
                self.oerp.db.drop(ARGS.super_admin_passwd, db)
            except:
                pass

# vim:expandtab:smartindent:tabstop=4:softtabstop=4:shiftwidth=4:



# -*- coding: UTF-8 -*-

try:
    import unittest2 as unittest
except ImportError:
    import unittest
import datetime

from args import ARGS

import oerplib
from oerplib.tools import v


class TestBrowse(unittest.TestCase):

    def setUp(self):
        self.oerp = oerplib.OERP(
            ARGS.server, protocol=ARGS.protocol, port=ARGS.port,
            version=ARGS.version)
        self.user = self.oerp.login(ARGS.user, ARGS.passwd, ARGS.database)

    def test_browse_with_one_id(self):
        # Check the result returned
        result = self.oerp.browse('res.users', self.user.id)
        self.assertIsInstance(result, oerplib.service.osv.browse.BrowseRecord)
        self.assertEqual(self.user, result)
        # With context
        context = self.oerp.execute('res.users', 'context_get')
        result = self.oerp.browse('res.users', self.user.id, context)
        self.assertIsInstance(result, oerplib.service.osv.browse.BrowseRecord)
        self.assertEqual(self.user, result)

    def test_browse_with_ids(self):
        # Iteration
        for result in self.oerp.browse('res.users', [self.user.id]):
            self.assertEqual(self.user, result)
        user_ids = self.oerp.search('res.users', [])
        for result in self.oerp.browse('res.users', user_ids):
            self.assertIsInstance(
                result, oerplib.service.osv.browse.BrowseRecord)
        # With context
        context = self.oerp.execute('res.users', 'context_get')
        for result in self.oerp.browse('res.users', user_ids, context):
            self.assertIsInstance(
                result, oerplib.service.osv.browse.BrowseRecord)

    def test_browse_with_id_false(self):
        # Check the result returned
        result = self.oerp.browse('res.users', False)
        self.assertIsInstance(result, oerplib.service.osv.browse.BrowseRecord)
        self.assertEqual(False, result.id)
        # With context
        context = self.oerp.execute('res.users', 'context_get')
        result = self.oerp.browse('res.users', False, context)
        self.assertIsInstance(result, oerplib.service.osv.browse.BrowseRecord)
        self.assertEqual(False, result.id)

    def test_browse_with_wrong_id(self):
        # Handle exception (execute a 'browse' with wrong ID)
        self.assertRaises(
            oerplib.error.RPCError,
            self.oerp.browse,
            'res.users',
            999999999)

    def test_browse_without_args(self):
        # Handle exception (execute a 'browse' without args)
        self.assertRaises(
            TypeError,
            self.oerp.browse,
            'res.users')

    def test_browse_with_wrong_args(self):
        # Handle exception (execute a 'browse' with wrong args)
        self.assertRaises(
            oerplib.error.RPCError,
            self.oerp.browse,
            'res.users',
            "wrong_arg")  # Wrong arg

    def test_reset(self):
        # Check the result returned
        self.user.name = "Charly"
        self.oerp.reset(self.user)
        self.assertEqual(self.user.name, "Administrator")

    def test_refresh(self):
        # Check the result returned
        self.user.name = "Charly"
        self.oerp.refresh(self.user)
        self.assertEqual(self.user.name, "Administrator")

    def test_write_record_char(self):
        backup_name = self.user.name
        self.user.name = "Charly"
        self.oerp.write_record(self.user)
        self.assertEqual(self.user.name, "Charly")
        self.user.name = backup_name
        self.oerp.write_record(self.user)
        self.assertEqual(self.user.name, backup_name)

    def test_write_record_boolean(self):
        self.user.active = False
        self.user.active = True
        self.oerp.write_record(self.user)
        self.assertEqual(self.user.active, True)

    def test_write_record_float(self):
        partner = self.user.company_id.partner_id
        partner.credit_limit = False
        self.oerp.write_record(partner)
        self.assertEqual(partner.credit_limit, 0.0)
        partner.credit_limit = 0.0
        self.oerp.write_record(partner)
        self.assertEqual(partner.credit_limit, 0.0)

    def test_write_record_integer(self):
        cur = self.oerp.browse('res.currency', 1)
        backup_accuracy = cur.accuracy
        cur.accuracy = False
        self.oerp.write_record(cur)
        self.assertEqual(cur.accuracy, 0)
        cur.accuracy = backup_accuracy
        self.oerp.write_record(cur)
        self.assertEqual(cur.accuracy, backup_accuracy)

    def test_write_record_selection(self):
        self.user.context_tz = False
        self.oerp.write_record(self.user)
        self.assertEqual(self.user.context_tz, False)
        self.user.context_tz = 'Europe/Paris'
        self.oerp.write_record(self.user)
        self.assertEqual(self.user.context_tz, 'Europe/Paris')

    def test_write_record_date(self):
        partner = self.user.company_id.partner_id
        partner.date = False
        self.oerp.write_record(partner)
        self.assertEqual(partner.date, False)
        partner.date = '2012-01-01'
        self.oerp.write_record(partner)
        self.assertEqual(partner.date, datetime.date(2012, 1, 1))
        partner.date = datetime.date(2012, 1, 1)
        self.oerp.write_record(partner)
        self.assertEqual(partner.date, datetime.date(2012, 1, 1))

    #def test_write_record_datetime(self):
    #    # No common model found in every versions of OpenERP with a
    #    # fields.datetime writable
    #    pass

    def test_write_record_many2many(self):
        backup_groups = list(self.user.groups_id)
        # False
        self.user.groups_id = False
        self.assertEqual(list(self.user.groups_id), [])
        self.oerp.reset(self.user)
        # []
        self.user.groups_id = []
        self.assertEqual(list(self.user.groups_id), [])
        self.oerp.reset(self.user)
        # [(6, 0, IDS)]
        self.user.groups_id = [(6, 0, [1, 2])]
        self.assertIn(1, [grp.id for grp in self.user.groups_id])
        self.assertIn(2, [grp.id for grp in self.user.groups_id])
        self.oerp.write_record(self.user)
        self.assertIn(1, [grp.id for grp in self.user.groups_id])
        self.assertIn(2, [grp.id for grp in self.user.groups_id])
        # Operator +=
        grp_obj = self.oerp.get('res.groups')
        self.user.groups_id += 4                        # ID
        self.assertIn(4, [grp.id for grp in self.user.groups_id])
        self.oerp.reset(self.user)
        self.user.groups_id += grp_obj.browse(4)        # Browse record
        self.assertIn(4, [grp.id for grp in self.user.groups_id])
        self.oerp.reset(self.user)
        self.user.groups_id += [4, 5]                   # List of IDs
        group_ids = [grp.id for grp in self.user.groups_id]
        self.assertIn(4, group_ids)
        self.assertIn(5, group_ids)
        self.oerp.reset(self.user)
        self.user.groups_id += list(grp_obj.browse([4, 5]))  # List of browse records
        group_ids = [grp.id for grp in self.user.groups_id]
        self.assertIn(4, group_ids)
        self.assertIn(5, group_ids)
        self.oerp.write_record(self.user)
        self.assertIn(4, group_ids)
        self.assertIn(5, group_ids)
        self.user.groups_id = False                     # Starting with no value
        self.user.groups_id += 4
        self.user.groups_id += [5]
        self.assertIn(4, [grp.id for grp in self.user.groups_id])
        self.assertIn(5, [grp.id for grp in self.user.groups_id])
        # Operator -=
        self.user.groups_id -= 1                        # ID
        self.assertNotIn(1, [grp.id for grp in self.user.groups_id])
        self.oerp.reset(self.user)
        self.user.groups_id -= grp_obj.browse(1)        # Browse record
        self.assertNotIn(1, [grp.id for grp in self.user.groups_id])
        self.oerp.reset(self.user)
        self.user.groups_id -= [1, 2]                   # List of IDs
        group_ids = [grp.id for grp in self.user.groups_id]
        self.assertNotIn(1, group_ids)
        self.assertNotIn(2, group_ids)
        self.oerp.reset(self.user)
        self.user.groups_id -= list(grp_obj.browse([1, 2]))  # List of browse records
        group_ids = [grp.id for grp in self.user.groups_id]
        self.assertNotIn(1, group_ids)
        self.assertNotIn(2, group_ids)
        self.oerp.write_record(self.user)
        self.assertNotIn(1, group_ids)
        self.assertNotIn(2, group_ids)
        self.user.groups_id = False                     # Starting with no value
        self.user.groups_id -= 1
        self.user.groups_id -= [5]
        self.assertEqual([], [grp.id for grp in self.user.groups_id])
        # Restore the original value
        self.user.groups_id = backup_groups
        self.assertEqual(list(self.user.groups_id), backup_groups)
        self.oerp.write_record(self.user)
        self.assertEqual(list(self.user.groups_id), backup_groups)

    def test_write_record_many2one(self):
        self.user.action_id = 1
        self.assertEqual(self.user.action_id.id, 1)
        self.oerp.write_record(self.user)
        self.assertEqual(self.user.action_id.id, 1)
        action = self.oerp.get('ir.actions.actions').browse(1)
        self.user.action_id = action
        self.oerp.write_record(self.user)
        self.assertEqual(self.user.action_id.id, 1)
        # False
        self.user.action_id = False
        self.oerp.write_record(self.user)
        if v(self.oerp.version)[:2] == v('5.0'):
            self.assertEqual(self.user.action_id.id, 1)
        elif v(self.oerp.version)[:2] == v('6.0'):
            self.assertEqual(self.user.action_id, False)
        elif v(self.oerp.version)[:2] == v('6.1'):
            self.assertEqual(self.user.action_id, False)
        else:
            self.assertEqual(self.user.action_id, False)

    def test_write_record_one2many(self):
        partner_obj = self.oerp.get('res.partner')
        partner = partner_obj.browse(1)
        backup_childs = [acc for acc in partner.child_ids]
        p1_id = partner_obj.create({'name': "Test 1"})
        p2_id = partner_obj.create({'name': "Test 2"})
        # False
        partner.child_ids = False
        self.assertEqual(list(partner.child_ids), [])
        self.oerp.write_record(partner)
        self.assertEqual(list(partner.child_ids), [])
        # []
        partner.child_ids = []
        self.assertEqual(list(partner.child_ids), [])
        self.oerp.write_record(partner)
        self.assertEqual(list(partner.child_ids), [])
        # [(6, 0, IDS)]
        partner.child_ids = [(6, 0, [p1_id])]
        self.assertEqual([acc.id for acc in partner.child_ids], [p1_id])
        self.oerp.write_record(partner)
        self.assertEqual([acc.id for acc in partner.child_ids], [p1_id])
        # Operator +=
        partner.child_ids += p2_id                      # ID
        self.assertIn(p2_id, [pt.id for pt in partner.child_ids])
        self.oerp.reset(partner)
        partner.child_ids += partner_obj.browse(p2_id)  # Browse record
        self.assertIn(p2_id, [pt.id for pt in partner.child_ids])
        self.oerp.reset(partner)
        partner.child_ids += [p1_id, p2_id]             # List of IDs
        partner_ids = [pt.id for pt in partner.child_ids]
        self.assertIn(p1_id, partner_ids)
        self.assertIn(p2_id, partner_ids)
        self.oerp.reset(partner)
        partner.child_ids += list(partner_obj.browse([p1_id, p2_id]))   # List of browse records
        partner_ids = [pt.id for pt in partner.child_ids]
        self.assertIn(p1_id, partner_ids)
        self.assertIn(p2_id, partner_ids)
        self.oerp.write_record(partner)
        self.assertIn(p1_id, partner_ids)
        self.assertIn(p2_id, partner_ids)
        partner.child_ids = False                       # Starting with no value
        partner.child_ids += p1_id
        partner.child_ids += [p2_id]
        self.assertIn(p1_id, [pt.id for pt in partner.child_ids])
        self.assertIn(p2_id, [pt.id for pt in partner.child_ids])
        # Operator -=
        partner.child_ids -= p1_id                      # ID
        self.assertNotIn(p1_id, [pt.id for pt in partner.child_ids])
        self.oerp.reset(partner)
        partner.child_ids -= partner_obj.browse(p1_id)  # Browse record
        self.assertNotIn(p1_id, [pt.id for pt in partner.child_ids])
        self.oerp.reset(partner)
        partner.child_ids -= [p1_id, p2_id]             # List of IDs
        partner_ids = [pt.id for pt in partner.child_ids]
        self.assertNotIn(p1_id, partner_ids)
        self.assertNotIn(p2_id, partner_ids)
        self.oerp.reset(partner)
        partner.child_ids -= list(partner_obj.browse([p1_id, p2_id]))   # List of browse records
        partner_ids = [pt.id for pt in partner.child_ids]
        self.assertNotIn(p1_id, partner_ids)
        self.assertNotIn(p2_id, partner_ids)
        self.oerp.write_record(partner)
        self.assertNotIn(p1_id, partner_ids)
        self.assertNotIn(p2_id, partner_ids)
        partner.child_ids = False                       # Starting with no value
        partner.child_ids -= p1_id
        partner.child_ids -= [p2_id]
        self.assertEqual([], [pt.id for pt in partner.child_ids])
        # Restore the original value
        partner.child_ids = backup_childs
        self.assertEqual(list(partner.child_ids), backup_childs)
        self.oerp.write_record(partner)
        self.assertEqual(list(partner.child_ids), backup_childs)

# vim:expandtab:smartindent:tabstop=4:softtabstop=4:shiftwidth=4:
